/*
 * Object definitions relevant to analyzer.
 */

include("AStar.js");

const nodeType = {
	NONE: 0,
	STARTPOS: 1,
	DERRICK_FREE: 2,
	DERRICK_CLAIMED: 3,
	DERRICK_BUILT: 4,
	GATEWAY: 5
}

function Edge(otherCoords, distance) {
	this.to = new Point(otherCoords.r, otherCoords.c);
	this.dist = distance;
}

Edge.prototype = {
	isEqual: function(other) {
		if (other == null) {
			return false;
		}
		if (defined(other.to)) {
			return other.to.r === this.r && other.to.c === this.c;
		}
		return other.r === this.r && other.c === this.c;
	}
}

function GraphNode(coord, width, type) {
	this.coords = coord;
	this.edges = {};
	this.type = type;
	this.areaWidth = width;
	this.fortified = 0;
	this.lastSeen = 0;
	this.owner = null;
	this.isChokepoint = false;
	this.hazard = 0;
};

GraphNode.prototype = {
	addEdge: function(otherCoords, distance) {
		if (defined(this.edges[otherCoords.toString()])) { return; }
		this.edges[otherCoords.toString()] = new Edge(otherCoords, distance);
	}
}

function GateNode(start, end, coord, width, type) {
	GraphNode.call(this, coord, width, type);

	this.start = start;
	this.end = end;
	this.isChokepoint = true;
}

GateNode.prototype = Object.create(GraphNode.prototype);
GateNode.prototype.constructor = GateNode;

function Graph() {
	this.nodes = new Matrix();
	this.allNodes = [];

	this.addEdge = function(node, otherCoords) {
		var dist = distance(node.coords, otherCoords);

		node.addEdge(otherCoords, dist);
		this.nodes.at(otherCoords).addEdge(node.coords, dist);
	}

	this.findNode = function(point) {
		for (var i = 0; i < Math.ceil(mapWidth > mapHeight ? mapWidth/2 : mapHeight/2); i++) {
			var source = new Point(point.r - i > 0 ? point.r - i : 0, point.c - i > 0 ? point.c - i : 0);
			var potential = [];
			for (var j = 0; j < 2*i + 1; j++) {
				if (defined(this.nodes.at(new Point(source.r, source.c + j))) && this.nodes.at(new Point(source.r, source.c + j)) != null) {
					potential.push(new Point(source.r, source.c + j));
				}
				if ((source.r + j > 0 && source.r + j < mapHeight) && defined(this.nodes.at(new Point(source.r + j, source.c))) && this.nodes.at(new Point(source.r + j, source.c)) != null) {
					potential.push(new Point(source.r + j, source.c ));
				}
				if ((source.r + 2*i > 0 && source.r + 2*i < mapHeight) && defined(this.nodes.at(new Point(source.r + 2*i, source.c + j))) && this.nodes.at(new Point(source.r + 2*i, source.c + j)) != null) {
					potential.push(new Point(source.r + 2*i, source.c + j));
				}
				if ((source.r + j > 0 && source.r + j < mapHeight) && defined(this.nodes.at(new Point(source.r + j, source.c + 2*i))) && this.nodes.at(new Point(source.r + j, source.c + 2*i)) != null) {
					potential.push(new Point(source.r + j, source.c + 2*i));
				}
			}

			if (!potential.length) { continue; }

			var min = potential[0];
			for (var j = 1; j < potential.length; j++) {
				if (distance(point, potential[j]) < distance(point, min)) {
					min = potential[j];
				}
			}

			return this.nodes.at(min);
		}
		// unreachable (theoretically)
		throw "Couldn't find graph node " + point.toString();
	}

	this.nearestNode = function(source, predicate, propulsion, /*o*/amount) {
		return profile('AStarConditioned', source, predicate, propulsion, amount);
	}

	this.findPath = function(from, to, propulsion) {
		return profile('AStar', from, to, propulsion);
	}

	/*
	 * Prints graph in graphviz format, engine "neato"
	 */
	this.printGraph = function() {
		log("graph {");

		for (var i = 0; i < mapHeight - 1; i++) {
			for (var j = 0; j < mapWidth - 1; j++) {
				var p = new Point(i,j);
				var n = this.nodes.at(p);
				if (!n || n.edges.length == 0) { continue; }

				log("\""
					+ p.toString()
					+ "\""
					+ " [pos=\""
					+ JSON.stringify(n.coords.c)
					+ ",-"
					+ JSON.stringify(n.coords.r)
					+ "!\""
					+ ((n.type === nodeType.NONE) ? ", fillcolor=black" : (
						n.type === nodeType.STARTPOS ? ", fillcolor=green" : (
						n.type === nodeType.GATEWAY ? ", fillcolor=red" : ", fillcolor=blue")))
					+ ", style=filled, label=\"\"]");

				for (var k in n.edges) {
					log("\""
					+ p.toString()
					+ "\" -- "
					+ "\""
					+ k
					+ "\""
					// + " [label=\""
					// + n.edges[k].dist
					+ "[color=\""
					+ (n.edges[k].water ? "blue\"]" : "black\"]"));
				}
			}
		}
		log("}");
	}
}

function Matrix() {
	this.arr = new Array(mapHeight);

	for (var i = 0; i < this.arr.length; i++) {
		this.arr[i] = new Array(mapWidth);
	}

	this.at = function(point) {
		return this.arr[point.r][point.c];
	}
	this.set = function(point, value) {
		this.arr[point.r][point.c] = value;
	}
	// returns value of parameter's  bottom neighbour
	this.valueDown = function(point) {
		return this.arr[point.r + 1][point.c];
	}
	this.valueUp = function(point) {
		return this.arr[point.r - 1][point.c];
	}
	this.valueLeft = function(point) {
		return this.arr[point.r][point.c - 1];
	}
	this.valueRight = function(point) {
		return this.arr[point.r][point.c + 1];
	}
}