/*
 * This script analyzes map for tactical use. Source: /cite{thinning}
 */

include("/multiplay/skirmish/analyzerObjects.js");
include("/multiplay/skirmish/graphMaker.js");

/*
 * Determines if all points are reachable by given propulsion.
 */
function allNodesReachableBy(nodes, prop) {
	for (var i = 0, l = nodes.length; i < l; i++) {
		if (!propulsionCanReach(prop, startPositions[me].x, startPositions[me].y, nodes[i].x, nodes[i].y)) {
			return false;
		}
	}
	return true;
}

/*
 * Determines general topology of a map.
 */
function mapType() {
	if (allNodesReachableBy(derrickPositions.concat(startPositions), propulsion.Wheels.id)) {
		return 'land';
	}

	if (allNodesReachableBy(derrickPositions, propulsion.Hover.id)) {
		// only need hover for some derricks
		if (allNodesReachableBy(startPositions, propulsion.Wheels.id)) {
			return 'semi-land';
		}

		// definitely need hover
		if (allNodesReachableBy(startPositions, propulsion.Hover.id)) {
			return 'water';
		}
	}

	return 'vtol';
}

function tilePassable(x, y, mode) {
	switch (mode) {
		case 'land': {
			return terrainType(x, y) !== TER_WATER && terrainType(x, y) !== TER_CLIFFFACE;
		}
		case 'semi-land':
		case 'water': {
			return terrainType(x, y) !== TER_CLIFFFACE;
		}
		case 'vtol': {
			return true;
		}
		case 'pathmaker': {
			return terrainType(x, y) !== TER_WATER;
		}
	}
}

/*
 * Fill map with 0s on reachable coordinates.
 */
function processMap(map, mode, start, end) {
	for (var r = start ? start.r : 0, l = end ? end.r : mapHeight; r < l; r++) {
		for (var c = start ? start.c : 0, k = end ? end.c : mapWidth; c < k; c++) {
			if (c <= 0 || c >= mapWidth - 1 || r <= 0 || r >= mapHeight - 1) {
				map[r][c] = 0;
				continue;
			}

			if (tilePassable(c, r, mode)) {
				map[r][c] = 1;
			} else {
				map[r][c] = 0;
			}
		}
	}
}

function isFurtherFromWall(map, r, c) {
	return map[r][c] !== 0 && map[r - 1][c] >= map[r][c] && map[r + 1][c] >= map[r][c] && map[r][c - 1] >= map[r][c] && map[r][c + 1] >= map[r][c];
}

/*
 * Assigns distance from closest edge to tiles.
 */
function evalMap(map) {
	var changed = true;

	while (changed) {
		changed = false;
		for (var i = 0; i < mapHeight; i++) {
			for (var j = 0; j < mapWidth; j++) {
				if (!defined(map[i][j])) {
					map[i][j] = 0;
					continue;
				}

				if (isFurtherFromWall(map, i, j)) {
					map[i][j]++;
					changed = true;
				}
			}
		}
	}
}

function makePOINodes(map, nodes) {
	// derrickPositions variable does not always contain ALL resource locations and enuming resources does not return pre-built derricks
	var oils = enumFeature(ALL_PLAYERS, 'OilResource').concat(enumArea(1, 1, mapWidth, mapHeight, ALL_PLAYERS, false).filter(function(s){return s.stattype === RESOURCE_EXTRACTOR;}));
	for (var i = 0; i < oils.length; i++) {
		var derrick = new Point(oils[i].y, oils[i].x);
		nodes[derrick.r][derrick.c] = 'd';
		graph.nodes.set(derrick, new GraphNode(derrick, map[derrick.r][derrick.c], nodeType.DERRICK_FREE));
	}

	for (var i = 0; i < startPositions.length; i++) {
		var base = new Point(startPositions[i].y, startPositions[i].x);

		nodes[base.r][base.c] = 's';
		graph.nodes.set(base, new GraphNode(base, map[base.r][base.c], nodeType.STARTPOS));
	}

	var gates = enumGateways();
	for (var i = 0; i < gates.length; i++) {
		var start = new Point(gates[i].y1, gates[i].x1);
		var end = new Point(gates[i].y2, gates[i].x2);
		var coords = new Point(Math.floor((end.r - start.r)/2) + start.r, Math.floor((end.c - start.c)/2) + start.c);

		nodes[coords.r][coords.c] = 'g';
		graph.nodes.set(coords, new GateNode(start, end, coords, map[coords.r][coords.c], nodeType.GATEWAY));
	}
}

/*
 * Return an array of neighbouring points
 */
function getNeighbours(map, r, c) {
	var L = new Array(0);

	if (isValidTile(map, r + 1, c)) {
		L.push(new Point(r + 1, c));
	}
	if (isValidTile(map, r - 1, c)) {
		L.push(new Point(r - 1, c));
	}
	if (isValidTile(map, r, c + 1)) {
		L.push(new Point(r, c + 1));
	}
	if (isValidTile(map, r, c - 1)) {
		L.push(new Point(r, c - 1));
	}
	return L;
}

function NZ(map, point) {
	var n = 0;
	var source = new Point(point.r - 1 > 0 ? point.r - 1 : 0, point.c - 1 > 0 ? point.c - 1 : 0);
	for (var j = 0; j < 2; j++) {
		if (map[source.r][source.c + j] !== 0) {
			n++;
		}
		if ((source.r + 2 > 0 && source.r + 2 < mapHeight) && map[source.r + 2][source.c + 2 - j] !== 0) {
			n++;
		}
		if ((source.r + 2 > 0 && source.r + 2 < mapHeight) && map[source.r + 2 - j][source.c] !== 0) {
			n++;
		}
		if ((source.r + j > 0 && source.r + j < mapHeight) && map[source.r + j][source.c + 2] !== 0) {
			n++;
		}
	}
	return n;
}

function Z0(map, point) {
	var c = 0;
	var n = [new Point(point.r - 1, point.c),
			new Point(point.r - 1, point.c - 1),
			new Point(point.r, point.c - 1),
			new Point(point.r + 1, point.c - 1),
			new Point(point.r + 1, point.c),
			new Point(point.r + 1, point.c + 1),
			new Point(point.r, point.c + 1),
			new Point(point.r - 1, point.c + 1),
			new Point(point.r - 1, point.c)];

	for (var i = 1; i < n.length; i++) {
		if (isValidTile(map, n[i].r, n[i].c) && n[i - 1].r >= 0 && (map[n[i - 1].r][n[i - 1].c] == 0 && map[n[i].r][n[i].c] != 0)) {
			c++;
		}
	}
	return c;
}

function thin(nodes, current, prev, point) {
	var nz = NZ(prev, point);
	if (!defined(nodes[point.r][point.c]) &&
		(nz >= 2 && nz <= 6) &&
		Z0(prev, point) == 1 &&
		((prev[point.r - 1][point.c] * prev[point.r][point.c - 1] * prev[point.r][point.c + 1]) == 0 || Z0(prev, new Point(point.r - 1, point.c)) != 1) &&
		((prev[point.r - 1][point.c] * prev[point.r][point.c - 1] * prev[point.r + 1][point.c]) == 0 || Z0(prev, new Point(point.r, point.c - 1)) != 1)) {
			current[point.r][point.c] = 0;
			return true;
	}

	return false;
}

function thinning(map, nodes, start, end) {
	var helperMap = new Array(map.length);

	for (var i = start ? (start.r - 1) : 0; i < map.length; i++) {
		helperMap[i] = new Array(map[i].length);
	}

	arrayCopy(map, helperMap);

	var changed = true;

	var prevMap = helperMap;
	var currMap = map;

	while (changed) {
		changed = false;
		for (var i = start ? start.r : 1; i < (end ? end.r : (mapHeight - 1)); i++) {
			for (var j = start ? start.c : 1; j < (end ? end.c : (mapWidth - 1)); j++) {
				if (currMap[i][j] > 0) {
					if (thin(nodes, currMap, prevMap, new Point(i, j)) && !changed) { changed = true; }
				}
			}
		}

		arrayCopy(currMap, prevMap);
	}
}

/*
 * Identifies nodes in map.
 */
function findNodes(map, nodes) {
	for (var i = 1; i < mapHeight - 1; i++) {
		for (var j = 1; j < mapWidth - 1; j++) {
			if (!defined(nodes[i][j]) && map[i][j] !== 0 && (NZ(map, new Point(i, j)) >= 3 || NZ(map, new Point(i, j)) <= 1)) {
				nodes[i][j] = 'p';
				graph.nodes.set(new Point(i, j), new GraphNode(new Point(i, j), map[i][j], nodeType.NONE));
			}
		}
	}
}

function createMentalMap() {
	var mapMatrix = new Array(mapHeight);

	for (var i = 0; i < mapHeight; i++) {
		mapMatrix[i] = new Array(mapWidth);
	}

	var nodeMatrix = new Array(mapHeight);

	for (var i = 0; i < mapHeight; i++) {
		nodeMatrix[i] = new Array(mapWidth);
	}

	processMap(mapMatrix, mapTop);
	// log(JSON.stringify(mapMatrix));
	makePOINodes(mapMatrix, nodeMatrix);
	thinning(mapMatrix, nodeMatrix);
	// log(JSON.stringify(mapMatrix));
	// log('+');
	findNodes(mapMatrix, nodeMatrix);
	// log(JSON.stringify(nodeMatrix));
	// log('*');
	// makeGraph(mapMatrix, nodeMatrix);
	profile('makeGraph', mapMatrix, nodeMatrix);
	graph.printGraph();
}