/*
 *	Definition of micro core base, common to all micro cores
 */

function MicroCore(group, id, role) {
	this.units = new group(id, role);
	this.intention = null;
	this.target = null;
	this.status = state.IDLE;
	this.path = [];
}

/*
 *	Recycles all units under its command.
 */
MicroCore.prototype.disband = function() {
	if (this.units.recycle() && !this.units.size()) {
		this.orderDone();
	}
}

/*
 *	Follows given path.
 */
MicroCore.prototype.followPath = function(path) {
	if (path.length) {
		if (distance(this.units.graphPos.coords, path[0]) < PRESENCE_RADIUS) {
			path.shift();
		}
	}

	if (!path.length) {
		this.orderDone();
	} else {
		this.units.move(path[0]);
	}
}

/*
 *	Issue order and target to the core.
 */
MicroCore.prototype.order = function(o, target) {
	if (this.status < state.ENGAGED) {
		if (this.target) {
			this.orderDone(this.target);
		}

		this.intention = o;
		this.target = target;

		if (colonel.targets[target]) {
			colonel.targets[target].force[this.units.units] = this.units.size();
		}
	}
}

MicroCore.prototype.orderDone = function(target) {
	this.status = state.IDLE;

	if (colonel.targets[target]) {
		delete colonel.targets[target].force[this.units.units];
	}

	this.intention = null;
	this.target = null;
}

/*
 *	Scouting routine.
 */
MicroCore.prototype.scout = function(target) {
	if (this.status >= state.ENGAGED) { return; }
		this.status = state.MOVING;
		this.units.move(target);
	if (distance(this.units.pos, target) < PRESENCE_RADIUS) {
		this.orderDone(target);
	}
}

/*
 *	Attack routine.
 */
MicroCore.prototype.attack = function(target, enemies) {
	var coord = defined(target.r) ? target : new Point(target.y, target.x);

	if (distance(this.units.pos, coord) < PRESENCE_RADIUS) {
		var prod = enemies.filter(isProduction).concat(enemies.filter(isGenerator));

		if (prod.length) {
			this.units.attack(prod[0]);
			this.status = state.ENGAGED;
		} else if (enemies.length && enemies[0].stattype !== OIL_RESOURCE) {
			this.units.attack(enemies[0]);
			this.status = state.ENGAGED;
		} else {
			this.orderDone(target);
		}

		return;
	}

	this.units.attackMove(target);
	this.status = state.MOVING;
}

/*
 *	Retreat routine.
 */
MicroCore.prototype.retreat = function() {
	var repfacs = enumStruct(me, struct['Repair Facility'].id).filter(isBuilt);
	var where = null;

	this.orderDone(this.target);

	if (repfacs.length) {
		where = graph.findNode(new Point(repfacs[0].y, repfacs[0].x));
	} else {
		where = graph.nearestNode(this.units.graphPos.coords, friendlyFortified, designer.templates[this.role].prop)[0];
	}

	if (where) {
		this.target = where.coords;
	} else {
		this.target = bases[me];
	}

	this.status = state.RETREAT;

	var path = graph.findPath(this.units.graphPos.coords, this.target, this.prop);

	if (path && path.length) {
		this.path = path;
		this.intention = 'followPath';

		this.followPath(this.path);
	}

	sendRequest('engi', 'Repair Facility');

	if (colonel.targets[this.target] && colonel.targets[this.target].task !== 'hold') {
		// we got trouble
		throw 'Attempting to retreat into a warzone';
	}

}

MicroCore.prototype.isUnderstrength = function() {
	return this.units.size() < MAX_SQUAD_SIZE;
}

MicroCore.prototype.isEmpty = function() {
	return !this.units.size();
}

MicroCore.prototype.queueProduction = function() {
	if (this.isUnderstrength()) {
		sendRequest('prod', this.role);
	}
}

/*
 *	Main process of the core.
 */
MicroCore.prototype.process = function() {
	if (!this.units.size()) { return; }

	if (this.type === 'vtol') {
		if (enumStruct(me, struct['VTOL Rearming Pad'].id).filter(function(s){return structureIdle(s);}).length < this.units.size()) {
			sendRequest('engi', 'VTOL Rearming Pad');
		}

		if (this.units.checkAmmo() < 15) {
			this.orderDone(this.target);
			this.intention = 'rearm';
			this.target = null;
			return;
		}
	}

	var e = this.units.see();
	switch (this.status) {
		case state.IDLE: {
			if (this.units.health() < REPAIR_THRESHOLD) {
				var repfacs = enumStruct(me, struct['Repair Facility'].id).filter(isBuilt);

				for (var i in repfacs) {
					if (distance(this.units.pos, new Point(repfacs[i].y, repfacs[i].x)) <= PRESENCE_RADIUS) {
						this.status = state.MOVING;
						this.units.repair(repfacs[i]);
						break;
					}
				}
			}
		}
		case state.MOVING: {
			if (this.intention) {
				this[this.intention](this.target, e);
			}
			break;
		}
		case state.ENGAGED: {
			if (!e.length) {
				this.orderDone(this.target);
				break;
			}

			var combatants = e.filter(isCombatant);

			if (combatants.length) {
				this.respond(combatants);
			} else {
				this.status = state.IDLE;

				if (this.intention) {
					this[this.intention](this.target, e);
				}
			}

			break;
		}
		case state.RETREAT: {
			if (this.path && this.path.length) {
				this.followPath(this.path);
			} else {
				if (distance(this.units.pos, this.target) < PRESENCE_RADIUS) {
					this.orderDone(this.target);
				}
			}
			break;
		}
	}
	if (focusedPlayer >= 0 && distance(this.units.pos, bases[focusedPlayer]) < PRESENCE_RADIUS && !e.length) {
		focusedPlayer = lockOnPlayer();
	}
}

/*
 *	Response to being attacked.
 */
MicroCore.prototype.respond = function(enemyForce) {
	var allies = enumRange(this.units.pos.c, this.units.pos.r, 12, ALLIES).filter(isCombatant);

	if (allies.length < enemyForce.length && distance(this.units.graphPos.coords, bases[me]) > 7) {
		var reinforce = colonel.sendAssistance(this.units, allies.length, enemyForce.length);

		if (!reinforce) {
			colonel.abortPlan(this.target);
			this.retreat();
			return;
		}
	}
	this.micro(enemyForce);
}

/*
 *	Micromanage individual units in combat.
 */
MicroCore.prototype.micro = function(enemies) {
	this.status = state.ENGAGED;
	if (!this.target) {
		this.target = this.units.pos;
	}

	var u = this.units.getUnits();

	for (var i = 0; i < u.length; i++) {
		var closest = enemies[0];
		var dist = distBetweenTwoPoints(u[i].x, u[i].y, enemies[0].x, enemies[0].y);
		var dmg = 0;
		var template = new Template('temp', idToName[u[i].body], idToName[u[i].propulsion], [idToName[u[i].weapons[0].id]], 'Weapon');

		for (var j = 0; j < enemies.length; j++) {
			if (dist > distBetweenTwoPoints(u[i].x, u[i].y, enemies[j].x, enemies[j].y) && template.damage(idToName[enemies[j].propulsion], idToName[enemies[j].body]) >= dmg) {
				closest = enemies[j];
				dist = distBetweenTwoPoints(u[i].x, u[i].y, enemies[j].x, enemies[j].y);
				dmg = template.damage(idToName[enemies[j].propulsion], idToName[enemies[j].body]);
			}
		}

		if (Stats.Weapon[idToName[u[i].weapons[0].id]].ShortRange/128 - 1 > dist) {
			var p = new Point(closest.y - u[i].y, closest.x - u[i].x);
			var dest = new Point(u[i].y - (p.r > 0 ? 2 : -2), u[i].x - (p.c > 0 ? 2 : -2));

			if (dest.r > 0 && dest.r < mapHeight && dest.c > 0 && dest.c < mapWidth) {
				orderDroidLoc(u[i], DORDER_MOVE, dest.c, dest.r);
			}
		} else {
			orderDroidObj(u[i], DORDER_ATTACK, closest);
		}
	}
}